print("import: 'theanolm'")
import theanolm

print("import: 'theanolm.commands'")
import theanolm.commands

print("import: 'theanolm.network'")
import theanolm.network

print("import: 'theanolm.parsing'")
import theanolm.parsing

print("import: 'theanolm.scoring'")
import theanolm.scoring

print("import: 'theanolm.training'")
import theanolm.training

print("import: 'theanolm.vocabulary'")
import theanolm.vocabulary

