from theanolm.backend import TextFileType
from theanolm.vocabulary.statistics import BigramStatistics
from wordclasses.theanobigramoptimizer import TheanoBigramOptimizer
from wordclasses.numpybigramoptimizer import NumpyBigramOptimizer
from wordclasses.wctool import main
