"""Modules implementing the theanolm subcommands.
"""

import theanolm.commands.train
import theanolm.commands.score
import theanolm.commands.decode
import theanolm.commands.sample
import theanolm.commands.version
